unit Iterator;

interface

uses
  Classes, DabEmployee;

type
  TCustomIterator = class(TObject)
  protected
    function GetCurrentItem : TObject; overload;
  public
    procedure First; virtual; abstract;
    procedure Next; virtual; abstract;
    function  IsDone : Boolean; virtual; abstract;

    // This is something of a problem. Cannot specify a more general type
    // The GoF talk about "polymorphic iteration" where the return type can
    // be specified when creating the iterator using templates in C++
    property CurrentItem : TObject read GetCurrentItem;
  end;


  // Need a concrete iterator for collection items
  TCollectionIterator = class(TCustomIterator)
  private
    // Note that iterator stores its own position in the list/aggregate
    // Can therefore have several iterators working on the same list at once
    FCurrentItemIndex : Integer;
    FCollection       : TCollection;
  protected
    function GetCurrentItem : TCollectionItem; overload;

    property CurrentItemIndex : Integer read FCurrentItemIndex write FCurrentItemIndex;
  public
    constructor Create(ACollection : TCollection);

    procedure First; override;
    procedure Next; override;
    function  IsDone : Boolean; override;

    // Note change of type
    property CurrentItem : TCollectionItem read GetCurrentItem;
  end;


  TProcessItemEvent = procedure (Sender : TObject;
                                 CurrItem : TDabEmployee;
                                 var Stop : Boolean) of object;

  // Can have a different one for employees
  TEmployeeIterator = class(TCollectionIterator)
  private
    FOnProcessItem : TProcessItemEvent;
  protected
    function GetCurrentItem : TDabEmployee; overload;
    // Internal iterator example named as per GoF
    function ProcessItem(Item : TDabEmployee;Output : TStrings) : Boolean; virtual;
  public
    constructor Create(ACollection : TDabEmployees);

    // Internal iterator example named as per GoF. Puts name and position of
    // all employees into Output
    procedure Traverse(Output : TStrings);

    // For different operations would need to
    //    1. Pass the operation to the Traverse method
    //    2. Subclass the iterator with different operation in each overridden
    //       ProcessItem method
    //    3. Define lots of Traverse type methods
    //    4. Use events
    procedure EventTraverse;

    // Note change of type
    property CurrentItem : TDabEmployee read GetCurrentItem;
  published
    property OnProcessItem : TProcessItemEvent read FOnProcessItem
                                               write FOnProcessItem;
  end;


  // Iterators by definition know about (ie are coupled to) the list/aggregate
  // they are designed for. However, there is no reason for a list to know
  // anything at all about iterators, especially if they are part of another
  // library, like all the list types in the VCL.
  // For this reason I don't like the GoF suggestion of a Factory Method in the
  // list constructor that creates an iterator (there are other reasons too)
  // Instead, use a Factory object. It is often a Singleton, although this
  // isn't essential
  TIteratorFactory = class(TObject)
  public
    constructor Create;

    function CreateIterator(const List : TObject) : TCustomIterator;
  end;

var
  // Singleton factory object
  IteratorFactory : TIteratorFactory;


type
  // My preferred factory objects often use class functions :
  TBetterIteratorFactory = class(TObject)
  public
    class function CreateIterator(const List : TObject) : TCustomIterator;
  end;


  // The traversal algorithm might be done in the list/aggregate, and the
  // iterator used only to hold state info. This is a cursor.

implementation

uses SysUtils;

{ TCustomIterator }

function TCustomIterator.GetCurrentItem : TObject;
begin
  Result := nil;
end;


{ TCollectionIterator }

constructor TCollectionIterator.Create(ACollection : TCollection);
begin
  inherited Create;
  FCollection := ACollection;
end;


function TCollectionIterator.GetCurrentItem : TCollectionItem;
begin
  // Let an exception be raised if necessary
  Result := FCollection.Items[CurrentItemIndex];
end;


procedure TCollectionIterator.First;
begin
  CurrentItemIndex := 0;
end;


procedure TCollectionIterator.Next;
begin
  CurrentItemIndex := CurrentItemIndex + 1;
end;


function TCollectionIterator.IsDone : Boolean;
begin
  Result := CurrentItemIndex >= FCollection.Count;
end;


{ TEmployeeIterator }

constructor TEmployeeIterator.Create(ACollection : TDabEmployees);
begin
  inherited Create(ACollection);
end;


function TEmployeeIterator.GetCurrentItem : TDabEmployee;
begin
  Result := TDabEmployee(inherited GetCurrentItem);
end;


function TEmployeeIterator.ProcessItem(Item : TDabEmployee;Output : TStrings) : Boolean;
begin
  // Always true in our example, but could be set to false to stop traversal
  Result := True;
  Output.Add(CurrentItem.FullName + ',  ' +
             CurrentItem.Position);
end;


procedure TEmployeeIterator.Traverse(Output : TStrings);
begin
  // This is how the GoF example works (sort of). I would tend to use
  // exceptions rather than boolean flags though, especially as on many
  // traversals there is never a reason to stop
  Output.Clear;
  First;

  while not IsDone do begin
    if not ProcessItem(CurrentItem,Output) then begin
      Break;
    end;

    Next;
  end;
end;


procedure TEmployeeIterator.EventTraverse;
  var
    Stop : Boolean;
begin
  // This is a modification of the GoF example with event handling to
  // provide the ProcessItem equivalent
  Stop := False;
  First;

  while not IsDone do begin
    if Assigned(FOnProcessItem) then begin
      FOnProcessItem(Self,CurrentItem,Stop);
    end;

    if Stop then begin
      Break;
    end;

    Next;
  end;
end;


{ TIteratorFactory }

var
  FactoryCount : Integer;

constructor TIteratorFactory.Create;
begin
  // There are a lot of different ways to do a singleton. This is one simple one
  if FactoryCount > 0 then begin
    raise Exception.Create('Only one Iterator Factory allowed');
  end;

  Inc(FactoryCount);
  inherited;
end;


function TIteratorFactory.CreateIterator(const List : TObject) : TCustomIterator;
begin
  if List is TDabEmployees then begin
    Result := TEmployeeIterator.Create(TDabEmployees(List));
  end else if List is TCollection then begin
    Result := TCollectionIterator.Create(TCollection(List));
  end else begin
    raise Exception.Create('No iterator defined for class ' + List.ClassName);
  end;
end;


{ TBetterIteratorFactory }

class function TBetterIteratorFactory.CreateIterator(const List : TObject) : TCustomIterator;
begin
  if List is TDabEmployees then begin
    Result := TEmployeeIterator.Create(TDabEmployees(List));
  end else if List is TCollection then begin
    Result := TCollectionIterator.Create(TCollection(List));
  end else begin
    raise Exception.Create('No iterator defined for class ' + List.ClassName);
  end;
end;


initialization
  FactoryCount    := 0;
  IteratorFactory := TIteratorFactory.Create;


finalization
  FreeAndNil(IteratorFactory);

end.
